mod esp;
pub(crate) mod battery;

use bq34z100::{Bq34z100g1Driver};

use crate::log::LogMessage;
use ds323x::{DateTimeAccess, Ds323x};
use esp_ota::mark_app_valid;

use eeprom24x::{Eeprom24x, Eeprom24xTrait, SlaveAddr};
use embedded_hal_bus::i2c::MutexDevice;
use embedded_svc::wifi::AccessPointInfo;

use esp_idf_hal::adc::oneshot::config::AdcChannelConfig;
use esp_idf_hal::adc::oneshot::{AdcChannelDriver, AdcDriver};
use esp_idf_hal::adc::{attenuation, Resolution, ADC1};
use esp_idf_hal::i2c::{APBTickType, I2cConfig, I2cDriver};
use esp_idf_hal::units::FromValueType;
use esp_idf_svc::eventloop::EspSystemEventLoop;
use esp_idf_svc::nvs::EspDefaultNvsPartition;
use esp_idf_svc::wifi::config::{ScanConfig, ScanType};
use esp_idf_svc::wifi::EspWifi;
use esp_idf_sys::esp_restart;
use esp_idf_sys::{
    esp_deep_sleep, esp_sleep_enable_ext1_wakeup,
    esp_sleep_ext1_wakeup_mode_t_ESP_EXT1_WAKEUP_ANY_LOW,
};
use once_cell::sync::Lazy;
use plant_ctrl2::sipo::ShiftRegister40;

use anyhow::{anyhow};
use anyhow::{bail, Ok, Result};
use serde::{Deserialize, Serialize};

use chrono::{DateTime, Utc};
use ds18b20::Ds18b20;
use std::result::Result::Ok as OkStd;
use std::sync::Mutex;
use std::time::Duration;

use crate::config::{BatteryBoardVersion, BoardVersion, PlantControllerConfig};
use crate::hal::battery::{print_battery_bq34z100, BatteryMonitor};
use crate::hal::esp::ESP;
use crate::hal::BoardHal::{Initial, V3, V4};
use crate::log::log;
use embedded_hal::digital::OutputPin;
use esp_idf_hal::delay::Delay;
use esp_idf_hal::gpio::{AnyInputPin, Gpio0, Gpio1, Gpio10, Gpio11, Gpio12, Gpio13, Gpio14, Gpio15, Gpio16, Gpio17, Gpio18, Gpio19, Gpio2, Gpio20, Gpio21, Gpio22, Gpio23, Gpio24, Gpio25, Gpio26, Gpio27, Gpio28, Gpio29, Gpio3, Gpio30, Gpio4, Gpio5, Gpio6, Gpio7, Gpio8, IOPin, InputOutput, Level, Output, PinDriver, Pull};
use esp_idf_hal::pcnt::{PcntChannel, PcntChannelConfig, PcntControlMode, PcntCountMode, PcntDriver, PinIndex, PCNT0};
use esp_idf_hal::prelude::Peripherals;
use esp_idf_hal::reset::ResetReason;
use esp_idf_svc::sntp::{self, SyncStatus};
use esp_idf_svc::systime::EspSystemTime;
use esp_idf_sys::{gpio_hold_dis, gpio_hold_en, vTaskDelay, EspError};
use one_wire_bus::OneWire;
use pca9535::{GPIOBank, Pca9535Immediate, StandardExpanderInterface};

//Only support for 8 right now!
pub const PLANT_COUNT: usize = 8;
const REPEAT_MOIST_MEASURE: usize = 1;

const TANK_MULTI_SAMPLE: usize = 11;

pub static I2C_DRIVER: Lazy<Mutex<I2cDriver<'static>>> = Lazy::new(PlantHal::create_i2c);

#[non_exhaustive]
struct V3Constants;

impl V3Constants {

    const PUMP8_BIT: usize = 0;
    const PUMP1_BIT: usize = 1;
    const PUMP2_BIT: usize = 2;
    const PUMP3_BIT: usize = 3;
    const PUMP4_BIT: usize = 4;
    const PUMP5_BIT: usize = 5;
    const PUMP6_BIT: usize = 6;
    const PUMP7_BIT: usize = 7;
    const MS_0: usize = 8;
    const MS_4: usize = 9;
    const MS_2: usize = 10;
    const MS_3: usize = 11;
    const MS_1: usize = 13;
    const SENSOR_ON: usize = 12;

    const SENSOR_A_1: u8 = 7;
    const SENSOR_A_2: u8 = 6;
    const SENSOR_A_3: u8 = 5;
    const SENSOR_A_4: u8 = 4;
    const SENSOR_A_5: u8 = 3;
    const SENSOR_A_6: u8 = 2;
    const SENSOR_A_7: u8 = 1;
    const SENSOR_A_8: u8 = 0;

    const SENSOR_B_1: u8 = 8;
    const SENSOR_B_2: u8 = 9;
    const SENSOR_B_3: u8 = 10;
    const SENSOR_B_4: u8 = 11;
    const SENSOR_B_5: u8 = 12;
    const SENSOR_B_6: u8 = 13;
    const SENSOR_B_7: u8 = 14;
    const SENSOR_B_8: u8 = 15;
}




const CHARGING: usize = 14;
const AWAKE: usize = 15;

const FAULT_3: usize = 16;
const FAULT_8: usize = 17;
const FAULT_7: usize = 18;
const FAULT_6: usize = 19;
const FAULT_5: usize = 20;
const FAULT_4: usize = 21;
const FAULT_1: usize = 22;
const FAULT_2: usize = 23;



const X25: crc::Crc<u16> = crc::Crc::<u16>::new(&crc::CRC_16_IBM_SDLC);

#[link_section = ".rtc.data"]
static mut LAST_WATERING_TIMESTAMP: [i64; PLANT_COUNT] = [0; PLANT_COUNT];
#[link_section = ".rtc.data"]
static mut CONSECUTIVE_WATERING_PLANT: [u32; PLANT_COUNT] = [0; PLANT_COUNT];
#[link_section = ".rtc.data"]
static mut LOW_VOLTAGE_DETECTED: bool = false;

#[link_section = ".rtc.data"]
static mut RESTART_TO_CONF: bool = false;

#[derive(Debug, PartialEq)]
pub enum Sensor {
    A,
    B,
}

pub struct PlantHal {}



pub struct HAL<'a>{
    pub config: PlantControllerConfig,
    pub board_hal: BoardHal<'a>,
    pub esp: ESP<'a>,
    pub battery_monitor: BatteryMonitor<'a>
}



pub enum BoardHal<'a>{
    Initial {
        general_fault: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>
    },
    V3 {
        shift_register: ShiftRegister40<
            PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>,
            PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>,
            PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>,
        >,
        shift_register_enable_invert:
            PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, esp_idf_hal::gpio::Output>,
        tank_channel: AdcChannelDriver<'a, Gpio5, AdcDriver<'a, esp_idf_hal::adc::ADC1>>,
        solar_is_day: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, esp_idf_hal::gpio::Input>,
        light: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>,
        main_pump: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>,
        tank_power: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>,
        general_fault: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>,
        signal_counter: PcntDriver<'a>,
        one_wire_bus: OneWire<PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>>,
        rtc:
            Ds323x<ds323x::interface::I2cInterface<MutexDevice<'a, I2cDriver<'a>>>, ds323x::ic::DS3231>,
        eeprom: Eeprom24x<
            MutexDevice<'a, I2cDriver<'a>>,
            eeprom24x::page_size::B32,
            eeprom24x::addr_size::TwoBytes,
            eeprom24x::unique_serial::No,
        >,
    },
    V4 {
        tank_channel: AdcChannelDriver<'a, Gpio5, AdcDriver<'a, esp_idf_hal::adc::ADC1>>,
        solar_is_day: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, esp_idf_hal::gpio::Input>,
        signal_counter: PcntDriver<'a>,
        charge_indicator: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>,
        awake: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, Output>,
        light: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>,
        tank_power: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>,
        one_wire_bus: OneWire<PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>>,
        rtc:
            Ds323x<ds323x::interface::I2cInterface<MutexDevice<'a, I2cDriver<'a>>>, ds323x::ic::DS3231>,
        eeprom: Eeprom24x<
            MutexDevice<'a, I2cDriver<'a>>,
            eeprom24x::page_size::B32,
            eeprom24x::addr_size::TwoBytes,
            eeprom24x::unique_serial::No,
        >,
        general_fault: PinDriver<'a, esp_idf_hal::gpio::AnyIOPin, InputOutput>,
        pump_expander: Pca9535Immediate<MutexDevice<'a, I2cDriver<'a>>>,
        sensor_expander: Pca9535Immediate<MutexDevice<'a, I2cDriver<'a>>>,
    }
}




#[derive(Serialize, Deserialize, PartialEq, Debug)]
pub struct BackupHeader {
    pub timestamp: i64,
    crc16: u16,
    pub size: usize,
}

impl BoardInteraction for HAL<'_> {
    fn set_charge_indicator(&mut self, charging: bool) {
        match &mut self.board_hal {
            V3 { shift_register, .. } => {
                shift_register.decompose()[CHARGING]
                    .set_state(charging.into())
                    .unwrap();
            }
            V4 { charge_indicator , ..} => {
                charge_indicator.set_state(charging.into()).unwrap();
            },
            Initial { ..  } => {

            }
        }
    }
    fn deep_sleep(&mut self, duration_in_ms: u64) -> ! {
        match &mut self.board_hal {
            V3 { shift_register, .. } => {
                shift_register.decompose()[AWAKE].set_low().unwrap();
            }
            V4 {  awake, .. } => {
                awake.set_low().unwrap();
            },
            Initial { .. } => {

            }
        }

        unsafe {
            //if we don't do this here, we might just revert newly flashed firmware
            mark_app_valid();
            //allow early wakeup by pressing the boot button
            if duration_in_ms == 0 {
                esp_restart();
            } else {
                //configure gpio 1 to wakeup on low, reused boot button for this
                esp_sleep_enable_ext1_wakeup(
                    0b10u64,
                    esp_sleep_ext1_wakeup_mode_t_ESP_EXT1_WAKEUP_ANY_LOW,
                );
                esp_deep_sleep(duration_in_ms);
            }
        };
    }
    fn get_backup_info(&mut self) -> Result<BackupHeader> {
        let eeprom = match &mut self.board_hal {
            V3 {  eeprom, .. } => {eeprom}
            V4 { eeprom, .. } => {eeprom },
            &mut Initial { .. } => {
                bail!("Board not configured yet")
            }
        };
        let dummy = BackupHeader {
            timestamp: 0,
            crc16: 0,
            size: 0,
        };
        let store = bincode::serialize(&dummy)?.len();
        let mut header_page_buffer = vec![0_u8; store];

        match eeprom.read_data(0, &mut header_page_buffer) {
            OkStd(_) => {}
            Err(err) => bail!("Error reading eeprom header {:?}", err),
        };
        println!("Raw header is {:?} with size {}", header_page_buffer, store);
        let header: BackupHeader = bincode::deserialize(&header_page_buffer)?;
        Ok(header)
    }
    fn get_backup_config(&mut self) -> Result<Vec<u8>> {
        let eeprom = match &mut self.board_hal {
            V3 {  eeprom, .. } => {eeprom}
            V4 { eeprom, .. } => {eeprom }
            &mut Initial { .. } => {
                bail!("Board not configured yet")
            }
        };

        let dummy = BackupHeader {
            timestamp: 0,
            crc16: 0,
            size: 0,
        };
        let store = bincode::serialize(&dummy)?.len();
        let mut header_page_buffer = vec![0_u8; store];

        match eeprom.read_data(0, &mut header_page_buffer) {
            OkStd(_) => {}
            Err(err) => bail!("Error reading eeprom header {:?}", err),
        };
        let header: BackupHeader = bincode::deserialize(&header_page_buffer)?;

        //skip page 0, used by the header
        let data_start_address = eeprom.page_size() as u32;
        let mut data_buffer = vec![0_u8; header.size];
        match eeprom.read_data(data_start_address, &mut data_buffer) {
            OkStd(_) => {}
            Err(err) => bail!("Error reading eeprom data {:?}", err),
        };

        let checksum = X25.checksum(&data_buffer);
        if checksum != header.crc16 {
            bail!(
                "Invalid checksum, got {} but expected {}",
                checksum,
                header.crc16
            );
        }

        Ok(data_buffer)
    }
    fn backup_config(&mut self, bytes: &[u8]) -> Result<()> {
        let time = self.get_rtc_time()?.timestamp_millis();
        let eeprom = match &mut self.board_hal {
            V3 {  eeprom, .. } => {
                eeprom
            }
            V4 { eeprom, .. } => { eeprom },
            &mut Initial { .. } => {
                bail!("Board not configured yet")
            }
        };

        let delay = Delay::new_default();

        let checksum = X25.checksum(bytes);
        let page_size = eeprom.page_size();

        let header = BackupHeader {
            crc16: checksum,
            timestamp: time,
            size: bytes.len(),
        };

        let encoded = bincode::serialize(&header)?;
        if encoded.len() > page_size {
            bail!(
                "Size limit reached header is {}, but firest page is only {}",
                encoded.len(),
                page_size
            )
        }
        let as_u8: &[u8] = &encoded;

        match eeprom.write_page(0, as_u8) {
            OkStd(_) => {}
            Err(err) => bail!("Error writing eeprom {:?}", err),
        };
        delay.delay_ms(5);

        let to_write = bytes.chunks(page_size);

        let mut lastiter = 0;
        let mut current_page = 1;
        for chunk in to_write {
            let address = current_page * page_size as u32;
            let eeprom = match &mut self.board_hal {
                V3 {  eeprom, .. } => {
                    eeprom
                }
                V4 { eeprom, .. } => { eeprom }
                &mut Initial { .. } => {
                    bail!("Board not configured yet")
                }
            };
            match eeprom.write_page(address, chunk) {
                OkStd(_) => {}
                Err(err) => bail!("Error writing eeprom {:?}", err),
            };
            current_page += 1;

            let iter = (current_page % 8) as usize;
            if iter != lastiter {
                for i in 0..PLANT_COUNT {
                    let _ = self.fault(i, iter == i);
                }
                lastiter = iter;
            }

            //update led here?
            delay.delay_ms(5);
        }
        Ok(())
    }
    fn is_day(& self) -> bool {
        match & self.board_hal {
            V3 {  solar_is_day, .. } => {solar_is_day.get_level().into()}
            V4 { solar_is_day, .. } => {solar_is_day.get_level().into() }
            Initial { .. } => {
                false
            }
        }
    }
    //should be multsampled
    fn water_temperature_c(&mut self) -> Result<f32> {
        let one_wire_bus = match &mut self.board_hal {
            V3 { one_wire_bus, .. } => one_wire_bus,
            V4 { one_wire_bus, .. } => one_wire_bus,
            &mut Initial { .. } => {
                bail!("Board not configured yet")
            }
        };

        one_wire_bus
            .reset(&mut self.esp.delay)
            .map_err(|err| -> anyhow::Error { anyhow!("Missing attribute: {:?}", err) })?;
        let first = one_wire_bus.devices(false, &mut self.esp.delay).next();
        if first.is_none() {
            bail!("Not found any one wire  Ds18b20");
        }
        let device_address = first
            .unwrap()
            .map_err(|err| -> anyhow::Error { anyhow!("Missing attribute: {:?}", err) })?;

        let water_temp_sensor = Ds18b20::new::<EspError>(device_address)
            .map_err(|err| -> anyhow::Error { anyhow!("Missing attribute: {:?}", err) })?;

        water_temp_sensor
            .start_temp_measurement(one_wire_bus, &mut self.esp.delay)
            .map_err(|err| -> anyhow::Error { anyhow!("Missing attribute: {:?}", err) })?;
        ds18b20::Resolution::Bits12.delay_for_measurement_time(&mut self.esp.delay);
        let sensor_data = water_temp_sensor
            .read_data(one_wire_bus, &mut self.esp.delay)
            .map_err(|err| -> anyhow::Error { anyhow!("Missing attribute: {:?}", err) })?;
        if sensor_data.temperature == 85_f32 {
            bail!("Ds18b20 dummy temperature returned");
        }
        Ok(sensor_data.temperature / 10_f32)
    }
    /// return median tank sensor value in milli volt
    fn tank_sensor_voltage(&mut self) -> Result<f32> {
        let (tank_power, tank_channel) = match &mut self.board_hal {
            V3 {  tank_power, tank_channel, .. } => {(tank_power,tank_channel)}
            V4 { tank_power, tank_channel, .. } => {(tank_power,tank_channel) }
            &mut Initial { .. } => {
                bail!("Board not configured yet")
            }
        };


        tank_power.set_high()?;
        //let stabilize
        self.esp.delay.delay_ms(100);

        let mut store = [0_u16; TANK_MULTI_SAMPLE];
        for multisample in 0..TANK_MULTI_SAMPLE {
            let value = tank_channel.read()?;
            store[multisample] = value;
        }
        tank_power.set_low()?;

        store.sort();
        let median_mv = store[6] as f32 / 1000_f32;
        Ok(median_mv)
    }
    fn set_low_voltage_in_cycle(&mut self) {
        unsafe {
            LOW_VOLTAGE_DETECTED = true;
        }
    }
    fn clear_low_voltage_in_cycle(&mut self) {
        unsafe {
            LOW_VOLTAGE_DETECTED = false;
        }
    }
    fn light(&mut self, enable: bool) -> Result<()> {
        let light = match &mut self.board_hal {
            V3 { light, .. } => light,
            V4 { light, .. } => light,
            &mut Initial { .. } => {
                bail!("Board not configured yet")
            }
        };

        unsafe { gpio_hold_dis(light.pin()) };
        light.set_state(enable.into())?;
        unsafe { gpio_hold_en(light.pin()) };
        Ok(())
    }
    fn pump(&mut self, plant: usize, enable: bool) -> Result<()> {
        match &mut self.board_hal {
            V3 { shift_register, .. } => {
                let index = match plant {
                    0 => V3Constants::PUMP1_BIT,
                    1 => V3Constants::PUMP2_BIT,
                    2 => V3Constants::PUMP3_BIT,
                    3 => V3Constants::PUMP4_BIT,
                    4 => V3Constants::PUMP5_BIT,
                    5 => V3Constants::PUMP6_BIT,
                    6 => V3Constants::PUMP7_BIT,
                    7 => V3Constants::PUMP8_BIT,
                    _ => bail!("Invalid pump {plant}",),
                };
                //currently infallible error, keep for future as result anyway
                shift_register.decompose()[index].set_state(enable.into())?;
            }
            V4 { pump_expander,  .. } => {
                if enable {
                    pump_expander.pin_set_high(GPIOBank::Bank0, plant.try_into()?)?;
                } else {
                    pump_expander.pin_set_low(GPIOBank::Bank0, plant.try_into()?)?;
                }
            },
            &mut Initial { .. } => {
                bail!("Board not configured yet")
            }
        }

        Ok(())
    }
    fn last_pump_time(&self, plant: usize) -> Option<DateTime<Utc>> {
        let ts = unsafe { LAST_WATERING_TIMESTAMP }[plant];
        DateTime::from_timestamp_millis(ts)
    }
    fn store_last_pump_time(&mut self, plant: usize, time: DateTime<Utc>) {
        unsafe {
            LAST_WATERING_TIMESTAMP[plant] = time.timestamp_millis();
        }
    }
    fn store_consecutive_pump_count(&mut self, plant: usize, count: u32) {
        unsafe {
            CONSECUTIVE_WATERING_PLANT[plant] = count;
        }
    }
    fn consecutive_pump_count(&mut self, plant: usize) -> u32 {
        unsafe { CONSECUTIVE_WATERING_PLANT[plant] }
    }
    fn fault(&mut self, plant: usize, enable: bool) -> Result<()>{
        match &mut self.board_hal {
            V3 { shift_register, .. } => {
                let index = match plant {
                    0 => FAULT_1,
                    1 => FAULT_2,
                    2 => FAULT_3,
                    3 => FAULT_4,
                    4 => FAULT_5,
                    5 => FAULT_6,
                    6 => FAULT_7,
                    7 => FAULT_8,
                    _ => panic!("Invalid plant id {}", plant),
                };
                shift_register.decompose()[index]
                    .set_state(enable.into())?;
            }
            V4 { pump_expander, .. } => {
                if enable {
                    pump_expander.pin_set_high(GPIOBank::Bank1, plant.try_into()?)?;
                } else {
                    pump_expander.pin_set_low(GPIOBank::Bank1, plant.try_into()?)?;
                }

            }
            &mut Initial { .. } => {
                bail!("Board not configured yet")
            }
        }
        Ok(())
    }
    fn low_voltage_in_cycle(&mut self) -> bool {
        unsafe { LOW_VOLTAGE_DETECTED }
    }
    fn any_pump(&mut self, enable: bool) -> Result<()> {
        match &mut self.board_hal {
            V3 { main_pump, .. } => {
                main_pump.set_state(enable.into())?;
            }
            V4 { .. } => {
                //does not exist in v4, ignore it
            }
            &mut Initial { .. } => {
                bail!("Board not configured yet")
            }
        }
        Ok(())

    }
    fn time(&mut self) -> Result<DateTime<Utc>> {
        let time = EspSystemTime {}.now().as_millis();
        let smaller_time = time as i64;
        let local_time = DateTime::from_timestamp_millis(smaller_time)
            .ok_or(anyhow!("could not convert timestamp"))?;
        Ok(local_time)
    }
    fn sntp(&mut self, max_wait_ms: u32) -> Result<DateTime<Utc>> {
        let sntp = sntp::EspSntp::new_default()?;
        let mut counter = 0;
        while sntp.get_sync_status() != SyncStatus::Completed {
            let delay = Delay::new_default();
            delay.delay_ms(100);
            counter += 100;
            if counter > max_wait_ms {
                bail!("Reached sntp timeout, aborting")
            }
        }

        self.time()
    }
    fn measure_moisture_hz(&mut self, plant: usize, sensor: Sensor) -> Result<f32> {
        match &mut self.board_hal {
            V3{  signal_counter, shift_register, .. } => {
                let mut results = [0_f32; REPEAT_MOIST_MEASURE];
                for repeat in 0..REPEAT_MOIST_MEASURE {
                    signal_counter.counter_pause()?;
                    signal_counter.counter_clear()?;
                    //Disable all
                    shift_register.decompose()[V3Constants::MS_4].set_high()?;

                    let sensor_channel = match sensor {
                        Sensor::A => match plant {
                            0 => V3Constants::SENSOR_A_1,
                            1 => V3Constants::SENSOR_A_2,
                            2 => V3Constants::SENSOR_A_3,
                            3 => V3Constants::SENSOR_A_4,
                            4 => V3Constants::SENSOR_A_5,
                            5 => V3Constants::SENSOR_A_6,
                            6 => V3Constants::SENSOR_A_7,
                            7 => V3Constants::SENSOR_A_8,
                            _ => bail!("Invalid plant id {}", plant),
                        },
                        Sensor::B => match plant {
                            0 => V3Constants::SENSOR_B_1,
                            1 => V3Constants::SENSOR_B_2,
                            2 => V3Constants::SENSOR_B_3,
                            3 => V3Constants::SENSOR_B_4,
                            4 => V3Constants::SENSOR_B_5,
                            5 => V3Constants::SENSOR_B_6,
                            6 => V3Constants::SENSOR_B_7,
                            7 => V3Constants::SENSOR_B_8,
                            _ => bail!("Invalid plant id {}", plant),
                        },
                    };


                    let is_bit_set = |b: u8| -> bool { sensor_channel & (1 << b) != 0 };
                    let pin_0 = &mut shift_register.decompose()[V3Constants::MS_0];
                    let pin_1 = &mut shift_register.decompose()[V3Constants::MS_1];
                    let pin_2 = &mut shift_register.decompose()[V3Constants::MS_2];
                    let pin_3 = &mut shift_register.decompose()[V3Constants::MS_3];
                    if is_bit_set(0) {
                        pin_0.set_high()?;
                    } else {
                        pin_0.set_low()?;
                    }
                    if is_bit_set(1) {
                        pin_1.set_high()?;
                    } else {
                        pin_1.set_low()?;
                    }
                    if is_bit_set(2) {
                        pin_2.set_high()?;
                    } else {
                        pin_2.set_low()?;
                    }
                    if is_bit_set(3) {
                        pin_3.set_high()?;
                    } else {
                        pin_3.set_low()?;
                    }

                    shift_register.decompose()[V3Constants::MS_4].set_low()?;
                    shift_register.decompose()[V3Constants::SENSOR_ON].set_high()?;

                    let delay = Delay::new_default();
                    let measurement = 100; // TODO what is this scaling factor? what is its purpose?
                    let factor = 1000f32 / measurement as f32;

                    //give some time to stabilize
                    delay.delay_ms(10);
                    signal_counter.counter_resume()?;
                    delay.delay_ms(measurement);
                    signal_counter.counter_pause()?;
                    shift_register.decompose()[V3Constants::MS_4].set_high()?;
                    shift_register.decompose()[V3Constants::SENSOR_ON].set_low()?;
                    delay.delay_ms(10);
                    let unscaled = signal_counter.get_counter_value()? as i32;
                    let hz = unscaled as f32 * factor;
                    log(
                        LogMessage::RawMeasure,
                        unscaled as u32,
                        hz as u32,
                        &plant.to_string(),
                        &format!("{sensor:?}"),
                    );
                    results[repeat] = hz;
                }
                results.sort_by(|a, b| a.partial_cmp(b).unwrap()); // floats don't seem to implement total_ord

                let mid = results.len() / 2;
                let median = results[mid];
                Ok(median)
            },
            V4 {sensor_expander, signal_counter, ..} => {
                let mut results = [0_f32; REPEAT_MOIST_MEASURE];
                for repeat in 0..REPEAT_MOIST_MEASURE {
                    signal_counter.counter_pause()?;
                    signal_counter.counter_clear()?;


                    const MS0: u8 = 1_u8;
                    const MS1: u8 = 0_u8;
                    const MS2: u8 = 3_u8;
                    const MS3: u8 = 4_u8;
                    const MS4: u8 = 2_u8;
                    const SENSOR_ON: u8 = 5_u8;
                    //Disable all
                    sensor_expander.pin_set_high(GPIOBank::Bank0, MS4)?;

                    let sensor_channel = match sensor {
                        Sensor::A =>{
                            plant as u32
                        },
                        Sensor::B => {
                            (15 - plant) as u32
                        },
                    };



                    let is_bit_set = |b: u8| -> bool { sensor_channel & (1 << b) != 0 };
                    if is_bit_set(0) {
                        sensor_expander.pin_set_high(GPIOBank::Bank0, MS0)?;
                    } else {
                        sensor_expander.pin_set_low(GPIOBank::Bank0, MS0)?;
                    }
                    if is_bit_set(1) {
                        sensor_expander.pin_set_high(GPIOBank::Bank0, MS1)?;
                    } else {
                        sensor_expander.pin_set_low(GPIOBank::Bank0, MS1)?;
                    }
                    if is_bit_set(2) {
                        sensor_expander.pin_set_high(GPIOBank::Bank0, MS2)?;
                    } else {
                        sensor_expander.pin_set_low(GPIOBank::Bank0, MS2)?;
                    }
                    if is_bit_set(3) {
                        sensor_expander.pin_set_high(GPIOBank::Bank0, MS3)?;
                    } else {
                        sensor_expander.pin_set_low(GPIOBank::Bank0, MS3)?;
                    }

                    sensor_expander.pin_set_low(GPIOBank::Bank0, MS4)?;
                    sensor_expander.pin_set_high(GPIOBank::Bank0, SENSOR_ON)?;

                    let delay = Delay::new_default();
                    let measurement = 100; // TODO what is this scaling factor? what is its purpose?
                    let factor = 1000f32 / measurement as f32;

                    //give some time to stabilize
                    delay.delay_ms(10);
                    signal_counter.counter_resume()?;
                    delay.delay_ms(measurement);
                    signal_counter.counter_pause()?;
                    sensor_expander.pin_set_high(GPIOBank::Bank0, MS4)?;
                    sensor_expander.pin_set_low(GPIOBank::Bank0, SENSOR_ON)?;
                    sensor_expander.pin_set_low(GPIOBank::Bank0, MS0)?;
                    sensor_expander.pin_set_low(GPIOBank::Bank0, MS1)?;
                    sensor_expander.pin_set_low(GPIOBank::Bank0, MS2)?;
                    sensor_expander.pin_set_low(GPIOBank::Bank0, MS3)?;
                    delay.delay_ms(10);
                    let unscaled = signal_counter.get_counter_value()? as i32;
                    let hz = unscaled as f32 * factor;
                    log(
                        LogMessage::RawMeasure,
                        unscaled as u32,
                        hz as u32,
                        &plant.to_string(),
                        &format!("{sensor:?}"),
                    );
                    results[repeat] = hz;
                }
                results.sort_by(|a, b| a.partial_cmp(b).unwrap()); // floats don't seem to implement total_ord

                let mid = results.len() / 2;
                let median = results[mid];
                Ok(median)
            }
            _ => {
                bail!("Not implemented for this board")
            }
        }

    }
    fn general_fault(&mut self, enable: bool) {
        let general_fault = match &mut self.board_hal {
            V3 {  general_fault, .. } => {general_fault}
            V4 { general_fault,  .. } => {general_fault},
            Initial { general_fault, .. } => {general_fault}
        };
        unsafe { gpio_hold_dis(general_fault.pin()) };
        general_fault.set_state(enable.into()).unwrap();
        unsafe { gpio_hold_en(general_fault.pin()) };
    }

    fn mode_override_pressed(&mut self) -> bool {
        self.esp.boot_button.get_level() == Level::Low
    }
    fn factory_reset(&mut self) -> Result<()> {
        println!("factory resetting");
        self.esp.delete_config()?;
        //destroy backup header
        let dummy: [u8; 0] = [];
        self.backup_config(&dummy)?;

        Ok(())
    }
    fn get_rtc_time(&mut self) -> Result<DateTime<Utc>> {
        let rtc = match &mut self.board_hal {
            V3 { rtc, .. } => {rtc}
            V4 { rtc, .. } => {rtc},
            Initial { .. } => {
                bail!("Board not configured yet")
            }
        };
        match rtc.datetime() {
            OkStd(rtc_time) => Ok(rtc_time.and_utc()),
            Err(err) => {
                bail!("Error getting rtc time {:?}", err)
            }
        }
    }
    fn set_rtc_time(&mut self, time: &DateTime<Utc>) -> Result<()> {
        let rtc = match &mut self.board_hal {
            V3 { rtc, .. } => {rtc}
            V4 { rtc, .. } => {rtc}
            Initial { .. } => {
                bail!("Board not configured yet")
            }
        };
        let naive_time = time.naive_utc();
        match rtc.set_datetime(&naive_time) {
            OkStd(_) => Ok(()),
            Err(err) => {
                bail!("Error getting rtc time {:?}", err)
            }
        }
    }
    fn wifi_scan(&mut self) -> Result<Vec<AccessPointInfo>> {
        self.esp.wifi_driver.start_scan(
            &ScanConfig {
                scan_type: ScanType::Passive(Duration::from_secs(5)),
                show_hidden: false,
                ..Default::default()
            },
            true,
        )?;
        Ok(self.esp.wifi_driver.get_scan_result()?)
    }
    fn test_pump(&mut self, plant: usize) -> Result<()> {
        self.any_pump(true)?;
        self.pump(plant, true)?;
        unsafe { vTaskDelay(30000) };
        self.pump(plant, false)?;
        self.any_pump(false)?;
        Ok(())
    }
    fn test(&mut self) -> Result<()> {
        self.general_fault(true);
        unsafe { vTaskDelay(100) };
        self.general_fault(false);
        unsafe { vTaskDelay(100) };
        self.any_pump(true)?;
        unsafe { vTaskDelay(500) };
        self.any_pump(false)?;
        unsafe { vTaskDelay(500) };
        self.light(true)?;
        unsafe { vTaskDelay(500) };
        self.light(false)?;
        unsafe { vTaskDelay(500) };
        for i in 0..PLANT_COUNT {
            self.fault(i, true)?;
            unsafe { vTaskDelay(500) };
            self.fault(i, false)?;
            unsafe { vTaskDelay(500) };
        }
        for i in 0..PLANT_COUNT {
            self.pump(i, true)?;
            unsafe { vTaskDelay(100) };
            self.pump(i, false)?;
            unsafe { vTaskDelay(100) };
        }
        for plant in 0..PLANT_COUNT {
            let a = self.measure_moisture_hz(plant, Sensor::A);
            let b = self.measure_moisture_hz(plant, Sensor::B);
            let aa = match a {
                OkStd(a) => a as u32,
                Err(_) => u32::MAX,
            };
            let bb = match b {
                OkStd(b) => b as u32,
                Err(_) => u32::MAX,
            };
            log(LogMessage::TestSensor, aa, bb, &plant.to_string(), "");
        }
        Delay::new_default().delay_ms(10);
        Ok(())
    }

    fn get_restart_to_conf(&mut self) -> bool {
        unsafe { RESTART_TO_CONF }
    }
    fn set_restart_to_conf(&mut self, to_conf: bool) {
        unsafe {
            RESTART_TO_CONF = to_conf;
        }
    }
}



pub trait BoardInteraction {
    fn set_charge_indicator(&mut self, charging: bool);
    fn deep_sleep(&mut self, duration_in_ms: u64) -> !;
    fn get_backup_info(&mut self) -> Result<BackupHeader>;
    fn get_backup_config(&mut self) -> Result<Vec<u8>>;
    fn backup_config(&mut self, bytes: &[u8]) -> Result<()>;
    fn is_day(&self) -> bool;
    //should be multsampled
    fn water_temperature_c(&mut self) -> Result<f32>;
    /// return median tank sensor value in milli volt
    fn tank_sensor_voltage(&mut self) -> Result<f32>;
    fn set_low_voltage_in_cycle(&mut self);
    fn clear_low_voltage_in_cycle(&mut self);
    fn light(&mut self, enable: bool) -> Result<()>;
    fn pump(&mut self, plant: usize, enable: bool) -> Result<()>;
    fn last_pump_time(&self, plant: usize) -> Option<DateTime<Utc>>;
    fn store_last_pump_time(&mut self, plant: usize, time: DateTime<Utc>);
    fn store_consecutive_pump_count(&mut self, plant: usize, count: u32);
    fn consecutive_pump_count(&mut self, plant: usize) -> u32;
    fn fault(&mut self, plant: usize, enable: bool) -> Result<()>;
    fn low_voltage_in_cycle(&mut self) -> bool;
    fn any_pump(&mut self, enable: bool) -> Result<()>;
    fn time(&mut self) -> Result<DateTime<Utc>>;
    fn sntp(&mut self, max_wait_ms: u32) -> Result<DateTime<Utc>>;
    fn measure_moisture_hz(&mut self, plant: usize, sensor: Sensor) -> Result<f32>;
    fn general_fault(&mut self, enable: bool);
    fn mode_override_pressed(&mut self) -> bool;
    fn factory_reset(&mut self) -> Result<()>;
    fn get_rtc_time(&mut self) -> Result<DateTime<Utc>>;
    fn set_rtc_time(&mut self, time: &DateTime<Utc>) -> Result<()>;
    fn wifi_scan(&mut self) -> Result<Vec<AccessPointInfo>>;
    fn test_pump(&mut self, plant: usize) -> Result<()>;
    fn test(&mut self) -> Result<()>;
    fn get_restart_to_conf(&mut self) -> bool;
    fn set_restart_to_conf(&mut self, to_conf: bool);

}



pub struct FreePeripherals {
    pub gpio0: Gpio0,
    pub gpio1: Gpio1,
    pub gpio2: Gpio2,
    pub gpio3: Gpio3,
    pub gpio4: Gpio4,
    pub gpio5: Gpio5,
    pub gpio6: Gpio6,
    pub gpio7: Gpio7,
    pub gpio8: Gpio8,
    //config button here
    pub gpio10: Gpio10,
    pub gpio11: Gpio11,
    pub gpio12: Gpio12,
    pub gpio13: Gpio13,
    pub gpio14: Gpio14,
    pub gpio15: Gpio15,
    pub gpio16: Gpio16,
    pub gpio17: Gpio17,
    pub gpio18: Gpio18,
    //i2c here
    pub gpio21: Gpio21,
    pub gpio22: Gpio22,
    pub gpio23: Gpio23,
    pub gpio24: Gpio24,
    pub gpio25: Gpio25,
    pub gpio26: Gpio26,
    pub gpio27: Gpio27,
    pub gpio28: Gpio28,
    pub gpio29: Gpio29,
    pub gpio30: Gpio30,
    pub pcnt0: PCNT0,
    pub adc1: ADC1,
}


impl PlantHal {
    fn create_i2c() -> Mutex<I2cDriver<'static>> {
        let peripherals = unsafe { Peripherals::new() };

        let config = I2cConfig::new()
            .scl_enable_pullup(true)
            .sda_enable_pullup(true)
            .baudrate(100_u32.kHz().into())
            .timeout(APBTickType::from(Duration::from_millis(100)));

        let i2c = peripherals.i2c0;
        let scl = peripherals.pins.gpio19.downgrade();
        let sda = peripherals.pins.gpio20.downgrade();

        Mutex::new(I2cDriver::new(i2c, sda, scl, &config).unwrap())
    }


    pub fn create()  -> Result<Mutex<HAL<'static>>> {
        let peripherals = Peripherals::take()?;
        let sys_loop = EspSystemEventLoop::take()?;
        let nvs = EspDefaultNvsPartition::take()?;
        let wifi_driver = EspWifi::new(peripherals.modem, sys_loop, Some(nvs))?;


        let mut boot_button = PinDriver::input(peripherals.pins.gpio9.downgrade())?;
        boot_button.set_pull(Pull::Floating)?;

        let free_pins = FreePeripherals {
            adc1: peripherals.adc1,
            pcnt0: peripherals.pcnt0,
            gpio0: peripherals.pins.gpio0,
            gpio1: peripherals.pins.gpio1,
            gpio2: peripherals.pins.gpio2,
            gpio3: peripherals.pins.gpio3,
            gpio4: peripherals.pins.gpio4,
            gpio5: peripherals.pins.gpio5,
            gpio6: peripherals.pins.gpio6,
            gpio7: peripherals.pins.gpio7,
            gpio8: peripherals.pins.gpio8,
            gpio10: peripherals.pins.gpio10,
            gpio11: peripherals.pins.gpio11,
            gpio12: peripherals.pins.gpio12,
            gpio13: peripherals.pins.gpio13,
            gpio14: peripherals.pins.gpio14,
            gpio15: peripherals.pins.gpio15,
            gpio16: peripherals.pins.gpio16,
            gpio17: peripherals.pins.gpio17,
            gpio18: peripherals.pins.gpio18,
            gpio21: peripherals.pins.gpio21,
            gpio22: peripherals.pins.gpio22,
            gpio23: peripherals.pins.gpio23,
            gpio24: peripherals.pins.gpio24,
            gpio25: peripherals.pins.gpio25,
            gpio26: peripherals.pins.gpio26,
            gpio27: peripherals.pins.gpio27,
            gpio28: peripherals.pins.gpio28,
            gpio29: peripherals.pins.gpio29,
            gpio30: peripherals.pins.gpio30,
        };

        let mut esp = ESP {
            mqtt_client: None,
            wifi_driver,
            boot_button,
            delay: Delay::new(1000)
        };


        //init,reset rtc memory depending on cause
        let mut init_rtc_store: bool = false;
        let mut to_config_mode: bool = false;
        let reasons = ResetReason::get();
        match reasons {
            ResetReason::Software => {}
            ResetReason::ExternalPin => {}
            ResetReason::Watchdog => {
                init_rtc_store = true;
            }
            ResetReason::Sdio => init_rtc_store = true,
            ResetReason::Panic => init_rtc_store = true,
            ResetReason::InterruptWatchdog => init_rtc_store = true,
            ResetReason::PowerOn => init_rtc_store = true,
            ResetReason::Unknown => init_rtc_store = true,
            ResetReason::Brownout => init_rtc_store = true,
            ResetReason::TaskWatchdog => init_rtc_store = true,
            ResetReason::DeepSleep => {}
            ResetReason::USBPeripheral => {
                init_rtc_store = true;
                to_config_mode = true;
            }
            ResetReason::JTAG => init_rtc_store = true,
        };
        log(
            LogMessage::ResetReason,
            init_rtc_store as u32,
            to_config_mode as u32,
            "",
            &format!("{reasons:?}"),
        );

        esp.init_rtc_deepsleep_memory(init_rtc_store, to_config_mode);
        let fs_mount_error = esp.mount_file_system().is_err();

        let config = esp.get_config();
        let hal = match config {
            Result::Ok(config) => {
                let board_hal : BoardHal = match config.hardware.board {
                    BoardVersion::INITIAL => {
                        let mut general_fault = PinDriver::input_output(free_pins.gpio6.downgrade())?;
                        general_fault.set_pull(Pull::Floating)?;
                        general_fault.set_low()?;

                        if fs_mount_error {
                            general_fault.set_high()?
                        }

                        Initial { general_fault}},
                    BoardVersion::V3 => {PlantHal::create_v3(free_pins)?},
                    BoardVersion::V4 => {PlantHal::create_v4(free_pins)?}
                };

                let battery_monitor : BatteryMonitor = match config.hardware.battery {
                    BatteryBoardVersion::Disabled => { BatteryMonitor::Disabled {}}
                    BatteryBoardVersion::BQ34Z100G1 => {
                        let mut battery_driver = Bq34z100g1Driver {
                            i2c: MutexDevice::new(&I2C_DRIVER),
                            delay: Delay::new(0),
                            flash_block_data: [0; 32],
                        };
                        let status = print_battery_bq34z100(&mut battery_driver);
                        match status {
                            OkStd(_) => {}
                            Err(err) => {
                                log(
                                    LogMessage::BatteryCommunicationError,
                                    0u32,
                                    0,
                                    "",
                                    &format!("{err:?})"),
                                );
                            }
                        }
                        BatteryMonitor::BQ34Z100G1 { battery_driver }
                    }
                    BatteryBoardVersion::WchI2cSlave => {
                        BatteryMonitor::WchI2cSlave {}
                    }
                };


                HAL{
                    config,
                    board_hal,
                    esp,
                    battery_monitor
                }
            }
            Err(err) => {
                log(
                    LogMessage::ConfigModeMissingConfig,
                    0,
                    0,
                    "",
                    &err.to_string(),
                );

                let mut general_fault = PinDriver::input_output(free_pins.gpio6.downgrade())?;
                general_fault.set_pull(Pull::Floating)?;
                general_fault.set_low()?;

                HAL{
                    config: Default::default(),
                    board_hal: Initial { general_fault},
                    esp,
                    battery_monitor: BatteryMonitor::Disabled {},
                }
            }
        };

        Ok(Mutex::new(hal))
    }

    fn create_v4(peripherals: FreePeripherals) -> Result<BoardHal<'static>> {
        let mut awake = PinDriver::output(peripherals.gpio15.downgrade())?;
        awake.set_high()?;

        let mut general_fault = PinDriver::input_output(peripherals.gpio6.downgrade())?;
        general_fault.set_pull(Pull::Floating)?;
        general_fault.set_low()?;

        println!("Init rtc driver");
        let mut rtc = Ds323x::new_ds3231(MutexDevice::new(&I2C_DRIVER));

        println!("Init rtc eeprom driver");
        let mut eeprom = {
            Eeprom24x::new_24x32(
                MutexDevice::new(&I2C_DRIVER),
                SlaveAddr::Alternative(true, true, true),
            )
        };

        let mut one_wire_pin = PinDriver::input_output_od(peripherals.gpio18.downgrade())?;
        one_wire_pin.set_pull(Pull::Floating)?;

        let one_wire_bus = OneWire::new(one_wire_pin)
            .map_err(|err| -> anyhow::Error { anyhow!("Missing attribute: {:?}", err) })?;


        let rtc_time = rtc.datetime();
        match rtc_time {
            OkStd(tt) => {
                println!("Rtc Module reports time at UTC {}", tt);
            }
            Err(err) => {
                println!("Rtc Module could not be read {:?}", err);
            }
        }
        match eeprom.read_byte(0) {
            OkStd(byte) => {
                println!("Read first byte with status {}", byte);
            }
            Err(err) => {
                println!("Eeprom could not read first byte {:?}", err);
            }
        }


        let mut signal_counter = PcntDriver::new(
            peripherals.pcnt0,
            Some(peripherals.gpio22),
            Option::<AnyInputPin>::None,
            Option::<AnyInputPin>::None,
            Option::<AnyInputPin>::None,
        )?;

        signal_counter.channel_config(
            PcntChannel::Channel0,
            PinIndex::Pin0,
            PinIndex::Pin1,
            &PcntChannelConfig {
                lctrl_mode: PcntControlMode::Keep,
                hctrl_mode: PcntControlMode::Keep,
                pos_mode: PcntCountMode::Increment,
                neg_mode: PcntCountMode::Hold,
                counter_h_lim: i16::MAX,
                counter_l_lim: 0,
            },
        )?;

        let adc_config = AdcChannelConfig {
            attenuation: attenuation::DB_11,
            resolution: Resolution::Resolution12Bit,
            calibration: esp_idf_hal::adc::oneshot::config::Calibration::Curve,
        };
        let tank_driver = AdcDriver::new(peripherals.adc1)?;
        let tank_channel: AdcChannelDriver<Gpio5, AdcDriver<esp_idf_hal::adc::ADC1>> =
            AdcChannelDriver::new(tank_driver, peripherals.gpio5, &adc_config)?;

        let mut solar_is_day = PinDriver::input(peripherals.gpio7.downgrade())?;
        solar_is_day.set_pull(Pull::Floating)?;

        let mut light = PinDriver::input_output(peripherals.gpio10.downgrade())?;
        light.set_pull(Pull::Floating)?;

        let mut tank_power = PinDriver::input_output(peripherals.gpio11.downgrade())?;
        tank_power.set_pull(Pull::Floating)?;

        let mut charge_indicator = PinDriver::input_output(peripherals.gpio3.downgrade())?;
        charge_indicator.set_pull(Pull::Floating)?;
        charge_indicator.set_low()?;


        let mut pump_expander = Pca9535Immediate::new(MutexDevice::new(&I2C_DRIVER), 32);

        //todo error handing if init error
        for pin in 0..8{
            let _ = pump_expander.pin_into_output(GPIOBank::Bank0, pin);
            let _ = pump_expander.pin_into_output(GPIOBank::Bank1, pin);
            let _ = pump_expander.pin_set_low(GPIOBank::Bank0, pin);
            let _ = pump_expander.pin_set_low(GPIOBank::Bank1, pin);
        }

        let mut sensor_expander = Pca9535Immediate::new(MutexDevice::new(&I2C_DRIVER), 34);
        for pin in 0..8{
            let _ = sensor_expander.pin_into_output(GPIOBank::Bank0, pin);
            let _ = sensor_expander.pin_into_output(GPIOBank::Bank1, pin);
            let _ = sensor_expander.pin_set_low(GPIOBank::Bank0, pin);
            let _ = sensor_expander.pin_set_low(GPIOBank::Bank1, pin);
        }

        Ok(V4 {
            awake,
            tank_channel,
            solar_is_day,
            signal_counter,
            light,
            tank_power,
            one_wire_bus,
            rtc,
            eeprom,
            general_fault,
            pump_expander,
            sensor_expander,
            charge_indicator
        })
    }

    fn create_v3(peripherals: FreePeripherals) -> Result<BoardHal<'static>> {
        let mut clock = PinDriver::input_output(peripherals.gpio15.downgrade())?;
        clock.set_pull(Pull::Floating)?;
        let mut latch = PinDriver::input_output(peripherals.gpio3.downgrade())?;
        latch.set_pull(Pull::Floating)?;
        let mut data = PinDriver::input_output(peripherals.gpio23.downgrade())?;
        data.set_pull(Pull::Floating)?;
        let shift_register = ShiftRegister40::new(clock, latch, data);
        //disable all
        for mut pin in shift_register.decompose() {
            pin.set_low()?;
        }

        let awake = &mut shift_register.decompose()[AWAKE];
        awake.set_high()?;

        let charging = &mut shift_register.decompose()[CHARGING];
        charging.set_high()?;

        let ms0 = &mut shift_register.decompose()[V3Constants::MS_0];
        ms0.set_low()?;
        let ms1 = &mut shift_register.decompose()[V3Constants::MS_1];
        ms1.set_low()?;
        let ms2 = &mut shift_register.decompose()[V3Constants::MS_2];
        ms2.set_low()?;
        let ms3 = &mut shift_register.decompose()[V3Constants::MS_3];
        ms3.set_low()?;

        let ms4 = &mut shift_register.decompose()[V3Constants::MS_4];
        ms4.set_high()?;

        println!("Init battery driver");


        println!("Init rtc driver");
        let mut rtc = Ds323x::new_ds3231(MutexDevice::new(&I2C_DRIVER));

        println!("Init rtc eeprom driver");
        let mut eeprom = {
            Eeprom24x::new_24x32(
                MutexDevice::new(&I2C_DRIVER),
                SlaveAddr::Alternative(true, true, true),
            )
        };

        let mut one_wire_pin = PinDriver::input_output_od(peripherals.gpio18.downgrade())?;
        one_wire_pin.set_pull(Pull::Floating)?;

        let rtc_time = rtc.datetime();
        match rtc_time {
            OkStd(tt) => {
                println!("Rtc Module reports time at UTC {}", tt);
            }
            Err(err) => {
                println!("Rtc Module could not be read {:?}", err);
            }
        }
        match eeprom.read_byte(0) {
            OkStd(byte) => {
                println!("Read first byte with status {}", byte);
            }
            Err(err) => {
                println!("Eeprom could not read first byte {:?}", err);
            }
        }


        let mut signal_counter = PcntDriver::new(
            peripherals.pcnt0,
            Some(peripherals.gpio22),
            Option::<AnyInputPin>::None,
            Option::<AnyInputPin>::None,
            Option::<AnyInputPin>::None,
        )?;

        signal_counter.channel_config(
            PcntChannel::Channel0,
            PinIndex::Pin0,
            PinIndex::Pin1,
            &PcntChannelConfig {
                lctrl_mode: PcntControlMode::Keep,
                hctrl_mode: PcntControlMode::Keep,
                pos_mode: PcntCountMode::Increment,
                neg_mode: PcntCountMode::Hold,
                counter_h_lim: i16::MAX,
                counter_l_lim: 0,
            },
        )?;



        let adc_config = AdcChannelConfig {
            attenuation: attenuation::DB_11,
            resolution: Resolution::Resolution12Bit,
            calibration: esp_idf_hal::adc::oneshot::config::Calibration::Curve,
        };
        let tank_driver = AdcDriver::new(peripherals.adc1)?;
        let tank_channel: AdcChannelDriver<Gpio5, AdcDriver<esp_idf_hal::adc::ADC1>> =
            AdcChannelDriver::new(tank_driver, peripherals.gpio5, &adc_config)?;

        let mut solar_is_day = PinDriver::input(peripherals.gpio7.downgrade())?;
        solar_is_day.set_pull(Pull::Floating)?;



        let mut light = PinDriver::input_output(peripherals.gpio10.downgrade())?;
        light.set_pull(Pull::Floating)?;

        let mut main_pump = PinDriver::input_output(peripherals.gpio2.downgrade())?;
        main_pump.set_pull(Pull::Floating)?;
        main_pump.set_low()?;
        let mut tank_power = PinDriver::input_output(peripherals.gpio11.downgrade())?;
        tank_power.set_pull(Pull::Floating)?;
        let mut general_fault = PinDriver::input_output(peripherals.gpio6.downgrade())?;
        general_fault.set_pull(Pull::Floating)?;
        general_fault.set_low()?;

        let one_wire_bus = OneWire::new(one_wire_pin)
            .map_err(|err| -> anyhow::Error { anyhow!("Missing attribute: {:?}", err) })?;


        let mut shift_register_enable_invert = PinDriver::output(peripherals.gpio21.downgrade())?;


        unsafe { gpio_hold_dis(shift_register_enable_invert.pin()) };
        shift_register_enable_invert.set_low()?;
        unsafe { gpio_hold_en(shift_register_enable_invert.pin()) };


        Ok(BoardHal::V3 {
            shift_register,
            shift_register_enable_invert,
            tank_channel,
            solar_is_day,
            light,
            main_pump,
            tank_power,
            general_fault,
            signal_counter,
            one_wire_bus,
            rtc,
            eeprom,
        })
    }
}
